"""
Task

This module contains actions related to the running task instance.
"""
import os

from typing import Dict, List, Union, Any, Optional

from ppa import task, appliance, ui, exceptions

from .records import PayloadKey

AnyJson = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]


def manifest() -> Dict[str, Any]:
    """Get the task manifest.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"
    A dictionary containing the task manifest.

??? example
    ```YAML linenums="1"
    - ppa.task.manifest:
      save: manifest
    ```
"""
    return task.manifest()


def payload() -> Dict[str, Any]:
    """Get the task payload.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"
    A dictionary containing the task payload.

??? example
    ```YAML linenums="1"
    - ppa.task.payload:
      save: payload
    ```
"""
    return task.payload()


def uuid() -> str:
    """Get the task UUID.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    A string containing the task UUID.

??? example
    ```YAML linenums="1"
    - ppa.task.uuid:
      save: uuid
    ```
"""
    return os.environ["PPA_URL"].split("/")[-1]


def url() -> str:
    """Get a URL used to open this task instance in the PPA interface.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    The task instance URL as a string.

??? example
    ```YAML linenums="1"
    - ppa.task.url:
      save: url
    ```
"""
    return f"https://{appliance.address()}/ui/activity/attach/?uuid={uuid()}"


def get_payload_values(keys: List[PayloadKey]) -> Dict[str, Any]:
    """Get specific keys & values from the task payload.

Only use this action if the task payload is a dictionary.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __keys:__ a list of [PayloadKeys](../inputs_outputs#payloadkey) to get from the payload

!!! tip "Missing & Optional Keys"

    Supplying a `default_value` to a [PayloadKey](../inputs_outputs#payloadkey) will assign a value if it is missing.

    This action will fail if any keys _without_ a `default_value` are missing from the payload.

??? success "Output"

    A dictionary containing the requested keys & values from the payload.

??? example "Example - Required & Optional Keys"
    ```YAML linenums="1"
    - ppa.task.get_payload_values:
        keys:
          - name: vpc_id
          - name: region_name
            default_value: null
      save: payload
    ```
"""
    if not all(["name" in key for key in keys]):
        message = "A payload key was supplied without a name."
        ui.output_error(message)
        raise ValueError(message)

    task_payload = payload()

    if not isinstance(task_payload, dict):
        message = "This action cannot retrieve values from the payload as it is not a dictionary."
        ui.output_error(message)
        raise TypeError(message)

    required_key_names = [key["name"] for key in keys if "default_value" not in key]
    if missing_required_keys := [key_name for key_name in required_key_names if key_name not in task_payload]:
        message = "The following required keys are missing from the task payload: '{}'.".format(", ".join(missing_required_keys))
        ui.output_error(message)
        raise KeyError(message)

    return {
        key["name"]: task_payload[key["name"]]
        if key["name"] in required_key_names
        else task_payload.get(key["name"], key["default_value"])
        for key in keys
    }


def user_password(text: str) -> str:
    """With the task operator's consent, get the password they used to authenticate to PPA.

This can be used to authenticate to other systems using the task operator's identity.

!!! question "Minimum Plugin Version: 5.1.0"

??? info "Input"

    - __text:__ the text to display when asking the user for consent

??? danger "Remember"

    This operation is not possible if SAML is used to authenticate users to PPA.

??? success "Output"
    The operator's password if the consent was given, or `null` if it was not given.

??? example
    ```YAML linenums="1"
    - ppa.task.user_password:
        text: PPA needs to use your login password to access a system
      save: operator_password
    ```
"""
    return task.user_password(text)


def started_by_user() -> bool:
    """Determine if a task was started by a PPA user.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by a user, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_user:
      save: interactive
    ```
"""
    return task.trigger().lower() == "ui"


def started_by_schedule() -> bool:
    """Determine if a task was started by a PPA schedule.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by a schedule, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_schedule:
      save: scheduled
    ```
"""
    return task.trigger().lower() == "job"


def started_by_api() -> bool:
    """Determine if a task was started by an API call.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    This action takes no inputs.

??? success "Output"

    `true` if the task was started by an API call, otherwise `false`.

??? example
    ```YAML linenums="1"
    - ppa.task.started_by_api:
      save: api_triggered
    ```
"""
    return task.trigger().lower() == "api"


def save_result(json: AnyJson, reason: Optional[str] = "Saving the task result") -> None:
    """Set the task JSON result.

A task's JSON result can be used to store information relevant to the run of a task.

This result is not shown in the PPA interface, but it can be fetched via the PPA API.

The action can only be run once per task, as the task result cannot be overwritten once it is set.

!!! question "Minimum Plugin Version: 4.1.0"

!!! danger "PPA Version"

    This action requires PPA v2.7.1 or newer.

??? info "Input"

    - __json:__ any valid JSON to store as the task result

    - __reason:__ text to display when the action runs (defaults to _Saving the task result_)

??? success "Output"

    Nothing is outputted by this action.

??? example

    The following action:

    ```YAML linenums="1"
    - ppa.task.save_result:
        json:
          message: The task completed successfully
    ```

    Will set the following task result:

    ```json
    {'message': 'The task completed successfully'}
    ```
"""
    try:
        task.set_result_json(json, reason=reason)
    except (exceptions.PPAVersionError, exceptions.ValueAlreadyExists) as e:
        ui.output_error(str(e))
        raise e
